#pragma once

#include <filesystem>
#include <J3ML/LinearAlgebra.hpp>
#include <fstream>
#include <iostream>
#include "glad/glad.h"

// LearnOpenGL::Shader
// OpenGL Shader Class Wrapper
// Updated by dawsh
// https://learnopengl.com/code_viewer_gh.php?code=includes/learnopengl/shader.h

namespace JGL {
    using namespace J3ML::LinearAlgebra;

    class Shader;
}

/// TODO: Eventually support Hot-reloading shaders like in shadertoy. Should help tremendously with debugging.

class JGL::Shader {
public:
    /// The default constructor does not initialize any member values.
    Shader() = default;

    /// Creates a shader by compiling a vertex and fragment program from the sources in the respective filesystem paths.
    Shader(std::filesystem::path vertex_source_path, std::filesystem::path fragment_source_path);

    /// Creates a shader by compiling a vertex and fragment program from the respective GLSL source-strings.
    Shader(const std::string& vertex_code, const std::string& fragment_code);

    /// @return True if the shader program successfully loaded and compiled.
    bool Loaded() const;
    /// @return The integer handle that OpenGL links to this shader program.
    unsigned int Handle() const;
    /// Enable this shader. All rendering performed thereafter will be affected by this shader code.
    /// @see UseDefault.
    void Use();

    static void Use(GLuint shader_program_id);
    static void UseDefault();

    // TODO: Implement for hot-reloading.
    void Reload();
    // TODO: Implement for hot-reloading.
    void Unload();

    /// @return The Uniform variable linked to the specified name.
    /// A Uniform is global a variable that is unique per shader program object, and can be accessed from any shader at any stage in the shader program.
    GLint Uniform(const std::string& name) const;

    /// @return The Attribute variable linked to the specified name.
    /// Attributes differ from Uniforms in that their value is different for each instance of the shader-program as it is running.
    GLint Attribute(const std::string& name) const;

    /// Sets a `uniform bool name = value` in the shader program.
    void SetBool (const std::string& name, bool value)    const;
    /// Sets a `uniform int name = value` in the shader program.
    void SetInt  (const std::string& name, int value)     const;
    /// Sets a `uniform float name = value` in the shader program.
    void SetFloat(const std::string& name, float value)   const;
    /// Sets a `uniform vec2 name = value` in the shader program.
    /// @note GLSL has builtin vec2, while we implement a Vector2 type. Please be aware there may be differences in implementation between them!
    /// @see class J3ML::LinearAlgebra::Vector2.
    void SetVector2(const std::string& name, const Vector2& value) const;
    /// Sets a `uniform vec2 name = value` in the shader program.
    /// @note GLSL has builtin vec2, while we implement a Vector2 type. Please be aware there may be differences in implementation between them!
    /// @see class J3ML::LinearAlgebra::Vector2.
    void SetVector2(const std::string& name, float x, float y) const;
    /// Sets a `uniform vec3 name = value` in the shader program.
    /// @note GLSL has builtin vec3, while we implement a Vector3 type. Please be aware there may be differences in implementation between them!
    /// @see class J3ML::LinearAlgebra::Vector3.
    void SetVector3(const std::string& name, const Vector3& value) const;
    /// Sets a `uniform vec3 name = value` in the shader program.
    /// @note GLSL has builtin vec3, while we implement a Vector3 type. Please be aware there may be differences in implementation between them!
    /// @see class J3ML::LinearAlgebra::Vector3.
    void SetVector3(const std::string& name, float x, float y, float z) const;
    /// Sets a `uniform vec4 name = value` in the shader program.
    /// @note GLSL has builtin vec4, while we implement aVector4 type. Please be aware there may be differences in implementation between them!
    /// @see class J3ML::LinearAlgebra::Vector4.
    void SetVector4(const std::string& name, const Vector4& value) const;
    /// Sets a `uniform vec4 name = value` in the shader program.
    /// @note GLSL has builtin vec4, while we implement a Vector4 type. Please be aware there may be differences in implementation between them!
    /// @see class J3ML::LinearAlgebra::Vector4.
    void SetVector4(const std::string& name, float x, float y, float z, float w) const;

    /// Sets a `uniform mat2 name = value` in the shader program.
    /// @note GLSL has builtin mat2, while we implement a Matrix2x2 type. Please be aware there may be differences in implementation between them!
    /// @see class J3ML::LinearAlgebra::Matrix2x2
    void SetMatrix2x2(const std::string& name, const Matrix2x2& value) const;

    /// Sets a `uniform mat3 name = value` in the shader program.
    /// @note GLSL has builtin mat3, while we implement a Matrix3x3 type. Please be aware there may be differences in implementation between them!
    /// @see class J3ML::LinearAlgebra::Matrix3x3
    void SetMatrix3x3(const std::string& name, const Matrix3x3& value) const;

    /// Sets a `uniform mat4 name = value` in the shader program.
    /// @note GLSL has builtin mat4, while we implement a Matrix4x4 type. Please be aware there may be differences in implementation between them!
    /// @see class J3ML::LinearAlgebra::Matrix4x4
    void SetMatrix4x4(const std::string& name, const Matrix4x4& value) const;


    // TODO: Implement Uniform-Setters for GLSL types that do not have a J3ML corollary (dmat3x4, dvec4, etc).

protected:
private:
    unsigned int id = 0;
    std::string vertexPath;
    std::string fragmentPath;
    static void checkCompileErrors(GLuint shader, const std::string& type);
};