#include <JGL/JGL.h>
#include <JGL/logger/logger.h>
#include <J3ML/Geometry/AABB.hpp>
#include <J3ML/Geometry/OBB.hpp>
#include <J3ML/Algorithm/Bezier.hpp>
#include <JGL/types/Light.h>
#include "internals/internals.h"


std::array<GLfloat, 16> JGL::OpenGLPerspectiveProjectionRH(float fovY, float aspect, float z_near, float z_far) {
    std::array<GLfloat, 16> result{};
    GLfloat f = 1.0f / std::tan(fovY * 0.5f * Math::Pi / 180.0f);
    result[0] = f / aspect;
    result[5] = f;
    result[10] = (z_far + z_near) / (z_near - z_far);
    result[11] = -1.0f;
    result[14] = (2.0f * z_far * z_near) / (z_near - z_far);
    return result;
}

void JGL::J3D::ChangeFOV(float fov) {
    JGL::J3D::fov = fov;
}

void JGL::J3D::ChangeFarPlane(float far_plane) {
    JGL::J3D::far_plane = far_plane;
}

void JGL::J3D::RequiredLight(const JGL::LightBase* light) {
    bool success = false;
    for (auto& i : current_state.required_lights)
        if (i == nullptr) {
            i = light;
            success = true;
            break;
        }

    if (!success)
        Logger::Error("You cannot specify more than 8 required lights.");
}

void JGL::J3D::OptionalLights(const JGL::LightBase** lights, const size_t& light_count) {
    for (size_t i = 0; i < light_count; i++)
        current_state.optional_lights.push_back(lights[i]);
}

void JGL::J3D::Begin(bool two_pass) {
    auto aspect = (float) window_size.x / (float) window_size.y;
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glLoadMatrixf(OpenGLPerspectiveProjectionRH(fov, aspect, 0.001, far_plane).data());
    glMatrixMode(GL_MODELVIEW);

    //Get what the draw color was before we did anything.
    glGetFloatv(GL_CURRENT_COLOR, current_state.draw_color);
    glColor4fv(current_state.draw_color);

    if (!glIsEnabled(GL_DEPTH_TEST))
        current_state.depth_test = false,
        glEnable(GL_DEPTH_TEST);
    else
        current_state.depth_test = true;

    current_state.vertex_array = false;
    if (!glIsEnabled(GL_VERTEX_ARRAY))
        current_state.vertex_array = false,
        glEnableClientState(GL_VERTEX_ARRAY);

    if (glIsEnabled(GL_NORMAL_ARRAY))
        current_state.normal_array = true,
        glDisableClientState(GL_NORMAL_ARRAY);

    if (glIsEnabled(GL_TEXTURE_COORD_ARRAY))
        current_state.texture_coordinate_array = true,
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);

    if (glIsEnabled(GL_TEXTURE_2D))
        current_state.texture_2D = true,
        glDisable(GL_TEXTURE_2D);

    current_state.cull_face = false;
    if (glIsEnabled(GL_CULL_FACE))
        current_state.cull_face = true,
        glDisable(GL_CULL_FACE);

    if (!glIsEnabled(GL_BLEND))
        current_state.blend = false,
        glEnable(GL_BLEND),
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    else
        current_state.blend = true;

    // Reset the lights.
    current_state.required_lights = empty_light_array;
    current_state.optional_lights = {};

    glDisable(GL_LIGHTING);
}

void JGL::J3D::End() {
    if (!current_state.depth_test)
        glDisable(GL_DEPTH_TEST);

    if (!current_state.vertex_array)
        glDisableClientState(GL_VERTEX_ARRAY);

    if (current_state.texture_2D)
        glEnable(GL_TEXTURE_2D);

    if (!current_state.blend)
        glDisable(GL_BLEND);

    if (current_state.cull_face)
        glEnable(GL_CULL_FACE);

    if (current_state.normal_array)
        glEnableClientState(GL_NORMAL_ARRAY);

    if (current_state.texture_coordinate_array)
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);

    if (glIsEnabled(GL_LIGHTING)) {
        glDisable(GL_LIGHTING);
        glDisable(GL_LIGHT0);
        glDisable(GL_LIGHT1);
        glDisable(GL_LIGHT2);
        glDisable(GL_LIGHT3);
        glDisable(GL_LIGHT4);
        glDisable(GL_LIGHT5);
        glDisable(GL_LIGHT6);
        glDisable(GL_LIGHT7);
    }

    current_state.required_lights = empty_light_array;
    current_state.optional_lights = {};

    //Put the draw color back how it was before.
    glColor4fv(current_state.draw_color);
}

void JGL::J3D::DrawLine(const Color4& color, const Vector3& A, const Vector3& B, float thickness) {
    Vector3 vertices[] = {A, B};

    glLineWidth(thickness);
    glColor4ubv(color.ptr());
    glVertexPointer(3, GL_FLOAT, sizeof(Vector3), vertices);
    glDrawArrays(GL_LINES, 0, 2);
    glColor4fv(current_state.draw_color);
}

void JGL::J3D::WireframeSphere(const Color4& color, const Vector3& position, float radius, float thickness, unsigned int sectors, unsigned int stacks) {
    Sphere sphere = {position, radius};
    BatchWireframeSphere(color,& sphere, 1, thickness, sectors, stacks);
}

void JGL::J3D::WireframeSphere(const Color4& color, const Sphere& sphere, float thickness, unsigned int sectors, unsigned int stacks) {
    BatchWireframeSphere(color,& sphere, 1, thickness, sectors, stacks);
}

void JGL::J3D::BatchWireframeSphere(const Color4& color, const Sphere* spheres, const size_t& sphere_count, float thickness, unsigned int sectors, unsigned int stacks) {
    // Create one sphere with a radius of 1 about 0, 0.
    float r = 1;
    std::vector<Vector3> vertices((sectors + 1) * (stacks + 1));

    int index = 0;
    for (int i = 0; i <= sectors; i++) {
        float lat = J3ML::Math::Pi * (-0.5 + (float) i / sectors);
        float z = J3ML::Math::Sin(lat);
        float zr = J3ML::Math::Cos(lat);

        for (int j = 0; j <= stacks; j++) {
            float lng = 2 * J3ML::Math::Pi * (float) (j - 1) / stacks;
            float x = J3ML::Math::Cos(lng);
            float y = J3ML::Math::Sin(lng);

            float pos_x = r * x * zr;
            float pos_y = r * y * zr;
            float pos_z = r * z;

            vertices[index++] = Vector3(pos_x, pos_y, pos_z);
        }
    }

    glLineWidth(thickness);
    glColor4ubv(color.ptr());
    glVertexPointer(3, GL_FLOAT, sizeof(Vector3), vertices.data());

    // Render each sphere in the batch at their given position and radius.
    for(size_t i = 0; i < sphere_count; i++) {
        glPushMatrix();
        glTranslatef(spheres[i].Position.x, spheres[i].Position.y, spheres[i].Position.z);
        glScalef(spheres[i].Radius, spheres[i].Radius, spheres[i].Radius);
        glDrawArrays(GL_LINE_LOOP, 0, vertices.size());
        glPopMatrix();
    }
    glColor4fv(current_state.draw_color);
}

void JGL::J3D::WireframeRevoSphere(const Color4& color, const Vector3& position, float radius, float thickness, unsigned int sectors, unsigned int revolutions, bool draw_stacks) {
    Sphere sphere = {position, radius};
    BatchWireframeRevoSphere(color,& sphere, 1, thickness, sectors, revolutions, draw_stacks);
}

void JGL::J3D::WireframeRevoSphere(const Color4& color, const Sphere& sphere, float thickness, unsigned int sectors, unsigned int revolutions, bool draw_stacks) {
    BatchWireframeRevoSphere(color,& sphere, 1, thickness, sectors, revolutions, draw_stacks);
}

void JGL::J3D::BatchWireframeRevoSphere(const Color4& color, const Sphere* spheres, const size_t& sphere_count, float thickness, unsigned int sectors, unsigned int revolutions, bool draw_stacks) {
    float r = 1;
    std::vector<Vector3> vertices;
    vertices.reserve((sectors + 1) * (revolutions + 1));

    std::vector<Vector3> cross_section(sectors + 1);
    for (int i = 0; i <= sectors; i++) {
        float lat = J3ML::Math::Pi * (-0.5 + (float)i / sectors);
        float z = J3ML::Math::Sin(lat);
        float zr = J3ML::Math::Cos(lat);
        cross_section[i] = Vector3(0, zr * r, z * r);
    }

    // Revolve
    for (int j = 0; j <= revolutions; j++) {
        float lng = 2 * J3ML::Math::Pi * (float)j / revolutions;
        float cosLng = J3ML::Math::Cos(lng);
        float sinLng = J3ML::Math::Sin(lng);

        for (const auto& point : cross_section) {
            float pos_x = point.y * cosLng;
            float pos_y = point.y * sinLng;
            float pos_z = point.z;

            vertices.emplace_back(pos_x, pos_y, pos_z);
        }
    }

    glLineWidth(thickness);
    glColor4ubv(color.ptr());

    VRamList vertex_data(vertices.data(), vertices.size());
    glBindBuffer(GL_ARRAY_BUFFER, vertex_data.GetHandle());
    glVertexPointer(3, GL_FLOAT, sizeof(Vector3), nullptr);

    // Render each sphere in the batch at their given position and radius.
    for(size_t i = 0; i < sphere_count; i++) {
        glPushMatrix();
        glTranslatef(spheres[i].Position.x, spheres[i].Position.y, spheres[i].Position.z);
        glScalef(spheres[i].Radius, spheres[i].Radius, spheres[i].Radius);
        glDrawArrays(GL_LINE_LOOP, 0, vertex_data.GetLength());
        if (draw_stacks)
            glRotatef(90, 0, 1, 0),
            glDrawArrays(GL_LINE_LOOP, 0, vertex_data.GetLength());
        glPopMatrix();
    }
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glColor4fv(current_state.draw_color);
}

void JGL::J3D::WireframeIcosphere(const Color4& color, const Vector3& position, float radius, float thickness, unsigned int subdivisions) {

    // NOTE2SELF: Code i'm borrowing this from uses float-packed-arrays rather than discrete Vectors
    // working on translating that correctly...

    // TODO: Revise this once J3ML::Geometry::Icosahedron is implemented.
    const float h_angle = J3ML::Math::Pi / 180.f * 72.f;    // 72 degree = 360 / 5;
    const float v_angle = J3ML::Math::Atan(1.0f / 2.f);     // elevation = 26.565;

    std::vector<Vector3> vertices;                   // array of 12 vertices (x,y,z)
    unsigned int index = 0;                                      // indices
    float z, xy;                                        // coords
    float hAngle1 = -Math::Pi / 2.f - h_angle / 2.f;    // start from -126 at 1st row
    float hAngle2 = -Math::Pi / 2.f;                    // start from -90 deg at 2nd row

    // the first top vertex at (0,0,r)
    vertices[0] = position + Vector3{0,0,radius};



    // compute 10 vertices at 1st and 2nd rows
    for (int i = 1; i <= 5; ++i)
    {

        z  = radius * Math::Sin(v_angle);   // elevation
        xy = radius * Math::Cos(v_angle);   // length on XY plane

        vertices[index++] = position + Vector3{xy * Math::Cos(hAngle1), xy * Math::Sin(hAngle1), z} ;
        vertices[index++] = position + Vector3{xy * Math::Cos(hAngle2), xy * Math::Sin(hAngle2), -z};

        // next horizontal angles
        hAngle1 += h_angle;
        hAngle2 += h_angle;
    }

    // the last bottom vertex at (0, 0, -r)
    vertices[11] = position + Vector3{0,0, -radius};

    //glLineWidth(thickness);
    //glColor4ubv(color.ptr());
    //glBindBuffer(GL_ARRAY_BUFFER, vertices.size());
    //glVertexPointer(3, GL_FLOAT, sizeof(Vector3), vertices.data());
    //glDrawArrays(GL_TRIANGLE_STRIP, 0, vertices.size());
    //glColor4fv(baseColor);

    std::vector<Vector3> tmpVertices;
    std::vector<Vector3> tmpIndices;
    std::vector<Vector2> newVs((subdivisions+1) * (subdivisions+2) / 2 * 3);
    const Vector3 v1, v2, v3;
    Vector3 newV1, newV2, newV3;

    int i, j, k;
    index = 0;              // unsigned int
    float a;                // lerp alpha
    unsigned int i1, i2;    // indices

    // copy prev arrays
    tmpVertices = vertices;
    //tmpIndices = indices;

    // iterate each triangle of icosahedron
    for (i = 0; i < tmpIndices.size(); i++)
    {
        //v1 = tmpVertices[tmpIndices[i]];
        //v2 = tmpVertices[tmpIndices[i+1]];
        //v3 = tmpVertices[tmpIndices[i+2]];

    }
}

void JGL::J3D::WireframeIcosahedron(const Color4& color, const Vector3& position, float radius, float thickness) {
    // TODO: Revise this once J3ML::Geometry::Icosahedron is implemented.
    const float h_angle = J3ML::Math::Pi / 180.f * 72.f;    // 72 degree = 360 / 5;
    const float v_angle = J3ML::Math::Atan(1.0f / 2.f);     // elevation = 26.565;

    std::array<Vector3, 12> vertices;                   // array of 12 vertices (x,y,z)
    int index = 0;                                      // indices
    float z, xy;                                        // coords
    float hAngle1 = -Math::Pi / 2.f - h_angle / 2.f;    // start from -126 at 1st row
    float hAngle2 = -Math::Pi / 2.f;                    // start from -90 deg at 2nd row

    // the first top vertex at (0,0,r)
    vertices[0] = position + Vector3{0,0,radius};



    // compute 10 vertices at 1st and 2nd rows
    for (int i = 1; i <= 5; ++i)
    {

        z  = radius * Math::Sin(v_angle);   // elevation
        xy = radius * Math::Cos(v_angle);   // length on XY plane

        vertices[index++] = position + Vector3{xy * Math::Cos(hAngle1), xy * Math::Sin(hAngle1), z} ;
        vertices[index++] = position + Vector3{xy * Math::Cos(hAngle2), xy * Math::Sin(hAngle2), -z};

        // next horizontal angles
        hAngle1 += h_angle;
        hAngle2 += h_angle;
    }

    // the last bottom vertex at (0, 0, -r)
    vertices[11] = position + Vector3{0,0, -radius};

    glLineWidth(thickness);
    glColor4ubv(color.ptr());
    //glBindBuffer(GL_ARRAY_BUFFER, vertices.size());
    glVertexPointer(3, GL_FLOAT, sizeof(Vector3), vertices.data());
    glDrawArrays(GL_TRIANGLE_STRIP, 0, vertices.size());
    glColor4fv(current_state.draw_color);
}

void JGL::J3D::BatchWireframeAABB(const Color4& color, const AABB* boxes, const size_t& box_count, float thickness) {
    glColor4ubv(color.ptr());
    glLineWidth(thickness);

    glBindBuffer(GL_ARRAY_BUFFER, ShapeCache::cube_vertex_data->GetHandle());
    glVertexPointer(3, GL_FLOAT, sizeof(Vector3), nullptr);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ShapeCache::cube_index_data->GetHandle());

    for (size_t i = 0; i < box_count; i++) {
        Vector3 delta = (boxes[i].maxPoint - boxes[i].minPoint) / 2;
        Vector3 center = boxes[i].Centroid();
        glPushMatrix();
        glTranslatef(center.x, center.y, center.z);
        glScalef(delta.x, delta.y, delta.z);
        glDrawElements(GL_LINE_LOOP, ShapeCache::cube_index_data->GetLength(), GL_UNSIGNED_INT, nullptr);
        glPopMatrix();
    }
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glColor4fv(current_state.draw_color);
}

void JGL::J3D::WireframeAABB(const Color4& color, const Vector3& pos, const Vector3& radii, float thickness) {
    AABB aabb = {Vector3(pos.x - radii.x, pos.y - radii.y, pos.z - radii.z), Vector3(pos.x + radii.x, pos.y + radii.y, pos.z + radii.z)};
    BatchWireframeAABB(color, &aabb, 1, thickness);
}

void JGL::J3D::BatchFillAABB(const Color4& color, const AABB* boxes, const size_t& box_count) {
    glColor4ubv(color.ptr());

    glBindBuffer(GL_ARRAY_BUFFER, ShapeCache::cube_vertex_data->GetHandle());
    glVertexPointer(3, GL_FLOAT, sizeof(Vector3), nullptr);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ShapeCache::cube_index_data->GetHandle());

    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);

    bool using_lights = false;
    if (UsingLighting()) {
        using_lights = true;
        glEnableClientState(GL_NORMAL_ARRAY);
        glBindBuffer(GL_ARRAY_BUFFER, ShapeCache::cube_normal_data->GetHandle());
        glNormalPointer(GL_FLOAT, sizeof(float), nullptr);
    }

    for (size_t i = 0; i < box_count; i++) {
        Vector3 delta = (boxes[i].maxPoint - boxes[i].minPoint) / 2;
        Vector3 center = boxes[i].Centroid();
        glPushMatrix();
        glTranslatef(center.x, center.y, center.z);
        glScalef(delta.x, delta.y, delta.z);
        if (using_lights)
            SelectLights(center);
        glDrawElements(GL_TRIANGLES, 36, GL_UNSIGNED_INT, nullptr);
        if (using_lights)
            ResetLights();

        glPopMatrix();
    }
    glColor4fv(current_state.draw_color);
    if (using_lights)
        glDisableClientState(GL_NORMAL_ARRAY);
    glDisable(GL_CULL_FACE);
}
void JGL::J3D::FillAABB(const Color4& color, const Vector3& pos, const Vector3& radii) {
    AABB box = {pos - radii, pos + radii};
    BatchFillAABB(color, &box, 1);
}

void JGL::J3D::FillAABB(const Color4& color, const AABB& aabb) {
    BatchFillAABB(color, &aabb, 1);
}

void JGL::J3D::FillSphere(const Color4& color, const Sphere& sphere, unsigned int sectors, unsigned int stacks) {
    BatchFillSphere(color, &sphere, 1, sectors, stacks);
}

void JGL::J3D::BatchFillSphere(const Color4& color, const Sphere* spheres, const size_t& sphere_count, unsigned int sectors, unsigned int stacks) {
    float r = 1;
    std::vector<Vector3> vertices((sectors + 1) * (stacks + 1));
    std::vector<unsigned int> indices; indices.reserve(sectors * stacks * 6);

    float two_pi = 2 * J3ML::Math::Pi;
    int index = 0;
    for (int i = 0; i <= sectors; i++) {
        float lat = J3ML::Math::Pi * (-0.5 + (float) i / sectors);
        float z = J3ML::Math::Sin(lat);
        float zr = J3ML::Math::Cos(lat);

        for (int j = 0; j <= stacks; j++) {
            float lng =  two_pi * (float) (j - 1) / stacks;
            float x = J3ML::Math::Cos(lng);
            float y = J3ML::Math::Sin(lng);

            float pos_x = r * x * zr;
            float pos_y = r * y * zr;
            float pos_z = r * z;
            vertices[index++] = Vector3(pos_x, pos_y, pos_z);
        }
    }

    for (int i = 0; i < sectors; i++) {
        for (int j = 0; j < stacks; j++) {
            int first_index = i * (stacks + 1) + j;
            int second_index = first_index + stacks + 1;

            indices.push_back(first_index);
            indices.push_back(second_index);
            indices.push_back(first_index + 1);

            indices.push_back(second_index);
            indices.push_back(second_index + 1);
            indices.push_back(first_index + 1);
        }
    }

    VRamList vertex_data(vertices.data(), vertices.size());
    VRamList index_data(indices.data(), indices.size());

    glColor4ubv(color.ptr());
    glBindBuffer(GL_ARRAY_BUFFER, vertex_data.GetHandle());
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, index_data.GetHandle());
    glVertexPointer(3, GL_FLOAT, sizeof(Vector3), nullptr);

    for (size_t i = 0; i < sphere_count; i++) {
        const Vector3 position = spheres[i].Position;
        glPushMatrix();
        glTranslatef(position.x, position.y, position.z);
        glScalef(spheres[i].Radius, spheres[i].Radius, spheres[i].Radius);
        glDrawElements(GL_TRIANGLES, index_data.GetLength(), GL_UNSIGNED_INT, nullptr);
        glPopMatrix();
    }
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    glColor4fv(current_state.draw_color);
}

void JGL::J3D::FillSphere(const Color4& color, const Vector3& position, float radius, unsigned int sectors, unsigned int stacks) {
    Sphere sphere = {position, radius};
    BatchFillSphere(color, &sphere, 1, sectors, stacks);
}

void JGL::J3D::WireframeAABB(const Color4& color, const AABB& aabb, float thickness) {
    BatchWireframeAABB(color, &aabb, 1, thickness);
}

// TODO Make it work the same as AABB batching. I couldn't get rotation to do anything more than twitch in place :(.
void JGL::J3D::BatchWireframeOBB(const Color4& color, const OBB* boxes, const size_t& box_count, float thickness) {
    std::array<Vector3, 8> corner_points;
    std::array<GLuint, 24> indices =
            {
                    0, 1, 1, 2, 2, 3, 3, 0,
                    4, 5, 5, 6, 6, 7, 7, 4,
                    0, 4, 1, 5, 2, 6, 3, 7
            };

    glLineWidth(thickness);
    glColor4ubv(color.ptr());
    glVertexPointer(3, GL_FLOAT, sizeof(Vector3), corner_points.data());

    for (size_t i = 0; i < box_count; i++) {
        boxes[i].GetCornerPoints(corner_points.data());
        glPushMatrix();
        glTranslatef(boxes[i].pos.x, boxes[i].pos.y, boxes[i].pos.z);
        glDrawElements(GL_LINES, indices.size(), GL_UNSIGNED_INT, indices.data());
        glPopMatrix();
    }
    glColor4fv(current_state.draw_color);
}

void JGL::J3D::WireframeOBB(const Color4& color, const OBB& obb, float thickness) {
    BatchWireframeOBB(color, &obb, 1, thickness);
}

/*
void JGL::J3D::WireframeOBB(const Color4& color, const Vector3& position, const Vector3& radii, const Matrix3x3& orientation, float thickness) {
    WireframeOBB(color, OBB(position, radii, orientation. * Vector3::UnitX, rotation * Vector3::UnitY, rotation * Vector3::UnitZ), thickness);
}
 */

void JGL::J3D::DrawCubicBezierCurve(const Color4& color, const Vector3& controlA, const Vector3& pointA,
                               const Vector3& pointB, const Vector3& controlB, int subdivisions, float thickness) {
    Vector3 last = controlA;
    const Vector3& first = controlB;
    for (int i = 0; i < subdivisions; ++i)
    {
        float alpha = (float) i / (float) subdivisions;
        Vector3 step = J3ML::Algorithm::Bezier(alpha, controlA, pointA, pointB, controlB);
        DrawLine(color, last, step, thickness);
        last = step;
    }

    // Have to manually draw the last segment of the curve.
    DrawLine(color, last, first, thickness);
}

void JGL::J3D::DrawVertexArray(const Color4& color, const VertexArray& vertex_array, const Vector3& position) {
    glColor4ubv(color.ptr());

    glEnableClientState(GL_VERTEX_ARRAY);
    glBindBuffer(GL_ARRAY_BUFFER, vertex_array.GetVertices()->GetHandle());
    glVertexPointer(3, GL_FLOAT, 0, nullptr);

    glPushMatrix();
    glTranslatef(position.x, position.y, position.z);
        //glScalef(1,1,1);
        glDrawArrays(GL_TRIANGLES, 0, vertex_array.GetVertices()->GetLength());
    glPopMatrix();

        //glDrawElements(GL_LINES, vertex_array.GetIndices()->GetLength(), GL_UNSIGNED_INT, nullptr);
        //std::cout << vertex_array.GetVertices()->GetLength() << std::endl;
    glColor4fv(current_state.draw_color);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

}