#include <JGL/types/Shader.h>

#include <glad/glad.h>

namespace JGL {
    Shader::Shader(std::filesystem::path vertex_source_path, std::filesystem::path fragment_source_path) {
        std::string vertex_code;
        std::string fragment_code;
        std::string geometry_code; // Currently unused, might be supported later.
        std::string compute_code;  // Currently unused, might be supported later.

        std::ifstream vShaderFile;
        std::ifstream fShaderFile;
        std::ifstream gShaderFile;
        std::ifstream cShaderFile;


        vShaderFile.exceptions(std::ifstream::failbit | std::ifstream::badbit);
        fShaderFile.exceptions(std::ifstream::failbit | std::ifstream::badbit);
        gShaderFile.exceptions(std::ifstream::failbit | std::ifstream::badbit);
        cShaderFile.exceptions(std::ifstream::failbit | std::ifstream::badbit);

        try {
            // Open Files
            vShaderFile.open(vertex_source_path);
            fShaderFile.open(fragment_source_path);

            std::stringstream vShaderStream, fShaderStream;

            // Read file's buffer contents into streams
            vShaderStream << vShaderFile.rdbuf();
            fShaderStream << fShaderFile.rdbuf();

            // close file handlers
            vShaderFile.close();
            fShaderFile.close();


            // convert stream into string

            vertex_code = vShaderStream.str();
            fragment_code = fShaderStream.str();

            if (false) {
                gShaderFile.open("");
                std::stringstream gShaderStream;
                gShaderStream << gShaderFile.rdbuf();
                gShaderFile.close();
                geometry_code = gShaderStream.str();
            }
        } catch (std::ifstream::failure& e) {
            std::cout << "ERROR::SHADER::FILE_NOT_SUCCESSFULLY_READ: " << e.what() << std::endl;
        }

        const char* vShaderCode = vertex_code.c_str();
        const char* fShaderCode = fragment_code.c_str();

        // 2. Compile shaders.
        unsigned int vertex, fragment;

        // vertex shader.
        vertex = glCreateShader(GL_VERTEX_SHADER);
        glShaderSource(vertex, 1, &vShaderCode, NULL);
        glCompileShader(vertex);
        checkCompileErrors(vertex, "VERTEX");

        // fragment shader
        fragment = glCreateShader(GL_FRAGMENT_SHADER);
        glShaderSource(fragment, 1, &fShaderCode, NULL);
        glCompileShader(fragment);
        checkCompileErrors(fragment, "FRAGMENT");

        // if geometry shader is given, compile geometry shader.
        unsigned int geometry;
        if (false) {
            // const char* gShaderCode = geometry_code.c_str();
            // geometry = glCreateShader(GL_GEOMETRY_SHADER);
            // glShaderSource(geometry, 1 &gShaderCode, NULL);
            // glCompileShader(geometry);
            // checkCompileErrors(geometry, "GEOMETRY");
        }

        // shader Program
        id = glCreateProgram();
        glAttachShader(id, vertex);
        glAttachShader(id, fragment);

        if (false)
            glAttachShader(id, geometry);

        glLinkProgram(id);
        checkCompileErrors(id, "PROGRAM");

        // delete the shaders as they're linked into our program now and are no longer necessary
        glDeleteShader(vertex);
        glDeleteShader(fragment);
        if (false)
            glDeleteShader(geometry);



    }

    void Shader::checkCompileErrors(GLuint shader, const std::string& type) {
        GLint success;
        GLchar infoLog[1024];

        if (type != "PROGRAM") {
            glGetShaderiv(shader, GL_COMPILE_STATUS, &success);
            if (!success) {
                glGetShaderInfoLog(shader, 1024, NULL, infoLog);
                std::cout << "ERROR::SHADER_COMPILATION_ERROR of " << type << "\n" << infoLog << std::endl;
            }
        } else {
            glGetProgramiv(shader, GL_LINK_STATUS, &success);
            if (!success) {
                glGetProgramInfoLog(shader, 1024, NULL, infoLog);
                std::cout << "ERROR::PROGRAM_LINKING_ERROR of " << type << "\n" << infoLog << std::endl;
            }
        }

        if (success)
            std::cout << "Shader compiled successfully" << std::endl;
    }

    GLint Shader::Uniform(const std::string &name) const {
        return glGetUniformLocation(id, name.c_str());
    }

    GLint Shader::Attribute(const std::string &name) const {
        return glGetAttribLocation(id, name.c_str());
    }

    Shader::Shader(const std::string &vertex_code, const std::string &fragment_code) {
        const char* vShaderCode = vertex_code.c_str();
        const char* fShaderCode = fragment_code.c_str();

        // 2. Compile shaders.
        unsigned int vertex, fragment;

        // vertex shader.
        vertex = glCreateShader(GL_VERTEX_SHADER);
        glShaderSource(vertex, 1, &vShaderCode, NULL);
        glCompileShader(vertex);
        checkCompileErrors(vertex, "VERTEX");

        // fragment shader
        fragment = glCreateShader(GL_FRAGMENT_SHADER);
        glShaderSource(fragment, 1, &fShaderCode, NULL);
        glCompileShader(fragment);
        checkCompileErrors(fragment, "FRAGMENT");

        // if geometry shader is given, compile geometry shader.
        unsigned int geometry;
        if (false) {
            // const char* gShaderCode = geometry_code.c_str();
            // geometry = glCreateShader(GL_GEOMETRY_SHADER);
            // glShaderSource(geometry, 1 &gShaderCode, NULL);
            // glCompileShader(geometry);
            // checkCompileErrors(geometry, "GEOMETRY");
        }

        // shader Program
        id = glCreateProgram();
        glAttachShader(id, vertex);
        glAttachShader(id, fragment);

        if (false)
            glAttachShader(id, geometry);

        glLinkProgram(id);
        checkCompileErrors(id, "PROGRAM");

        // delete the shaders as they're linked into our program now and are no longer necessary
        glDeleteShader(vertex);
        glDeleteShader(fragment);
        if (false)
            glDeleteShader(geometry);
    }

    void Shader::Use() {
        glUseProgram(id);
    }

    bool Shader::Loaded() const { return id != 0; }

    unsigned int Shader::Handle() const { return id;}

    void Shader::UseDefault() {
        glUseProgram(0);
    }

    void Shader::Use(GLuint shader_program_id) {
        glUseProgram(shader_program_id);
    }

    void Shader::SetBool(const std::string &name, bool value) const { glUniform1i(Uniform(name), (int)value); }

    void Shader::SetInt(const std::string &name, int value) const { glUniform1i(Uniform(name), value); }

    void Shader::SetFloat(const std::string &name, float value) const { glUniform1f(Uniform(name), value); }

    void Shader::SetVector2(const std::string &name, const Vector2 &value) const { glUniform2f(Uniform(name), value.x, value.y); }

    void Shader::SetVector2(const std::string &name, float x, float y) const { glUniform2f(Uniform(name), x, y); }

    void Shader::SetVector3(const std::string &name, const Vector3 &value) const { glUniform3f(Uniform(name), value.x, value.y, value.z); }

    void Shader::SetVector3(const std::string &name, float x, float y, float z) const { glUniform3f(Uniform(name), x, y, z); }

    void Shader::SetVector4(const std::string &name, const Vector4 &value) const { glUniform4f(Uniform(name), value.x, value.y, value.z, value.w); }

    void Shader::SetVector4(const std::string &name, float x, float y, float z, float w) const { glUniform4f(Uniform(name), x, y, z, w); }

    void Shader::SetMatrix2x2(const std::string &name, const Matrix2x2 &value) const {
        /// TODO: Verify if glsl expects row-major or col-major!!
        bool transpose = false;
        glUniformMatrix2fv(Uniform(name), 4, transpose, value.ptr());
    }

    void Shader::SetMatrix3x3(const std::string &name, const Matrix3x3 &value) const {
        /// TODO: Verify if glsl expects row-major or col-major!!
        bool transpose = false;
        glUniformMatrix3fv(Uniform(name), 9, transpose, value.ptr());
    }

    void Shader::SetMatrix4x4(const std::string &name, const Matrix4x4 &value) const {
        /// TODO: Verify if glsl expects row-major or col-major!!
        bool transpose = false;
        glUniformMatrix4fv(Uniform(name), 16, transpose, value.ptr());
    }
}